#!/usr/bin/env python
import re
import os
import subprocess
import random

# choose 
#cas = "maple" # "maxima"
cas = "maple"

# sorting functions
# http://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside
def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [ atoi(c) for c in re.split('(\d+)', text) ]

# seed nomalizer take the smaller alphabetic order
def normalize(s):
    s_r = s[::-1]
    ls   = s.split(',');   ls.sort(key=natural_keys);   s   = ','.join(ls)
    ls_r = s_r.split(','); ls_r.sort(key=natural_keys); s_r = ','.join(ls_r)
    if s < s_r:
        return s.replace(',','_')
    else:
        return s_r.replace(',','_')

# print seeds
def print_seedlist(sl,fo,val):
    fo.write('[')
    for i,e in enumerate(sl):
        fo.write("'"+str(e) +"'")
        if i < len(sl)-1:
            fo.write(",")
    fo.write('] '+str(val)+"\n")




# parse a file and extract the pareto according to the polynomial condition (0:BERNOULLI, 1:INTEGRAL, 2:REVERSE INTEGRAL)
def parse_ied_file_condition(filename_in,filename_out,l,condition):
    '''
    '''
    already_seen = set();

    vals  = []
    seeds = []
    polys = {}

    with open(filename_in,"r") as f:
        for line in f:
            m = re.search('^([01,]*)\s+[0-9.]+\s+[0-9.]+\s+[0-9.]+\s+(.*)$',line)
            seed = str(m.group(1))
            seed = normalize(seed)

            # check for already seen seeds ...
            if seed in already_seen:
                continue
            else:
                already_seen.add(seed)

            p = re.compile('([0-9]+,[0-9]+=[0-9]+);')
            coefs = p.findall(str(m.group(2)))

            # generate the current seed polynom
            if cas == "maxima":
                poly = "M_"+seed+"(p) := 0"
            elif cas == "maple":
                poly = "M_"+seed+" := p -> 0"
            else:
                os.error("maple or maxima undefined for \"cas\" parameter")

            for coef in coefs:
                m2 = re.search('^([0-9]+),1=([0-9]+)$',coef)
                if m2:
                    index = int(m2.group(1))
                    count = int(m2.group(2))
                    if condition == 0:
                        # 0 : BERNOULLI
                        poly += " + p^("+str(index)+")*(1-p)^("+str(l-index)+") * "+str(count)
                    else :
                        # 1 : INTEGRAL  or  2 : REVERSE INTEGRAL
                        if cas == "maxima":
                            poly += " + p^("+str(index+1)+") * sum(binomial("+str(l-index)+",k)/("+str(index+1)+"+k)*(-p)^k, k, 0, "+str(l-index)+") * "+str(count)
                        elif cas == "maple":
                            poly += " + p^("+str(index+1)+") * sum(binomial("+str(l-index)+",k)/("+str(index+1)+"+k)*(-p)^k, k=0.."+str(l-index)+") * "+str(count)
                        else:
                            os.error("maple or maxima undefined for \"cas\" parameter")
            poly += ";\n"

            # check dominance

            # (a.1) very first seed
            if vals == []:
                if cas == "maxima":
                    vals  = [('0.0','1.0')]
                elif cas == "maple":
                    vals  = [('0','1')]
                else:
                    os.error("maple or maxima undefined for \"cas\" parameter")
                seeds = [seed]
                polys = {seed : poly}
            else:

                # (a.2) any "current_seed" vs this "seed"
                ru = random.randint(1, 1000000)
                filename_ml = "/tmp/_f_"+seed+"_"+str(ru)+".ml"
                                
                with open(filename_ml,"w") as fo_ml :
                    if cas == "maxima":
                        header = "fpprec : 200;\n" + "fpprintprec : 30;\n" + "rootsepsilon : 1e-150;\n" + "DYNAMALLOC:TRUE;\n"
                    elif cas == "maple":
                        header = "Digits := 200;\n"
                    else:
                        os.error("maple or maxima undefined for \"cas\" parameter")


                    fo_ml.write(header + poly)
                    already_seed = set()
                    
                    for i in range(len(vals)):
                        current_val  = vals[i]
                        current_seed = seeds[i]
                        current_poly = polys[current_seed]

                        # if twice in the pareto : only one polynomial
                        if current_seed in already_seed:
                            continue
                        else:
                            already_seed.add(current_seed)
                            
                        # (a.2.1) print maxima for each seed "definition and comparison" vs this one
                        if condition < 2:
                            if cas == "maxima":
                                poly_compare = "M_"+current_seed+"_MINUS_"+seed+"(p) := M_"+current_seed+"(p) - M_"+seed+"(p);\n"
                            elif cas == "maple":
                                poly_compare = "M_"+current_seed+"_MINUS_"+seed+" := p -> M_"+current_seed+"(p) - M_"+seed+"(p);\n"
                            else:
                                os.error("maple or maxima undefined for \"cas\" parameter")
                        else:
                            if cas == "maxima":
                                poly_compare = "M_"+current_seed+"_MINUS_"+seed+"(p) := M_"+current_seed+"(1) - M_"+seed+"(1) + M_"+seed+"(p) - M_"+current_seed+"(p);\n"
                            elif cas == "maple":
                                poly_compare = "M_"+current_seed+"_MINUS_"+seed+":= p -> M_"+current_seed+"(1) - M_"+seed+"(1) + M_"+seed+"(p) - M_"+current_seed+"(p);\n"
                            else:
                                os.error("maple or maxima undefined for \"cas\" parameter")
                        fo_ml.write(current_poly);
                        fo_ml.write(poly_compare);

                        # (a.2.2) print maxima for each seed "dominance extraction" vs this one
                        if cas == "maxima":
                            dominance_extraction = \
                            "root_M_"+current_seed+"_MINUS_"+seed+" : bfloat(realroots(M_"+current_seed+"_MINUS_"+seed+"(p),1e-150));\n" + \
                            "root_M_"+current_seed+"_MINUS_"+seed+" : sort(root_M_"+current_seed+"_MINUS_"+seed+");\n" + \
                            "if (rhs(root_M_"+current_seed+"_MINUS_"+seed+"[1]) > 0) then root_M_"+current_seed+"_MINUS_"+seed+" : cons(x = 0,root_M_"+current_seed+"_MINUS_"+seed+");\n" + \
                            "if (rhs(root_M_"+current_seed+"_MINUS_"+seed+"[length(root_M_"+current_seed+"_MINUS_"+seed+")]) < 1) then root_M_"+current_seed+"_MINUS_"+seed+" : endcons(x = 1,root_M_"+current_seed+"_MINUS_"+seed+");\n" + \
                            "for i from 2 thru length(root_M_"+current_seed+"_MINUS_"+seed+") do (if (rhs(root_M_"+current_seed+"_MINUS_"+seed+"[i]) > 0) and (rhs(root_M_"+current_seed+"_MINUS_"+seed+"[i-1]) < 1) then (newline(),printf(true,\"M_"+current_seed+"_MINUS_"+seed+" ## [~h,~h] : ~d\",max(rhs(root_M_"+current_seed+"_MINUS_"+seed+"[i-1]),bfloat(0)),min(rhs(root_M_"+current_seed+"_MINUS_"+seed+"[i]),bfloat(1)),signum(M_"+current_seed+"_MINUS_"+seed+"((rhs(root_M_"+current_seed+"_MINUS_"+seed+"[i-1])+rhs(root_M_"+current_seed+"_MINUS_"+seed+"[i]))/2))),newline()));\n"
                        elif cas == "maple":
                            dominance_extraction = \
                            "root_M_"+current_seed+"_MINUS_"+seed+" := [fsolve(M_"+current_seed+"_MINUS_"+seed+"(p),p=0..1,fulldigits)];\n" \
                            "if (root_M_"+current_seed+"_MINUS_"+seed+"[1] > 0) then\n" \
                            "  root_M_"+current_seed+"_MINUS_"+seed+" := [0 , op(root_M_"+current_seed+"_MINUS_"+seed+")];\n" \
                            "end if;\n" \
                            "if (root_M_"+current_seed+"_MINUS_"+seed+"[nops(root_M_"+current_seed+"_MINUS_"+seed+")] < 1) then\n" \
                            "  root_M_"+current_seed+"_MINUS_"+seed+" := [op(root_M_"+current_seed+"_MINUS_"+seed+") , 1];\n" \
                            "end if;\n" \
                            "for i from 2 to nops(root_M_"+current_seed+"_MINUS_"+seed+") do\n" \
                            "  if (root_M_"+current_seed+"_MINUS_"+seed+"[i] > 0) and (root_M_"+current_seed+"_MINUS_"+seed+"[i-1] < 1) then\n"  \
                            "    printf(\"M_"+current_seed+"_MINUS_"+seed+" ## [%.30G,%.30G] : %d\\n\",max(root_M_"+current_seed+"_MINUS_"+seed+"[i-1],0),min(root_M_"+current_seed+"_MINUS_"+seed+"[i],1),signum(M_"+current_seed+"_MINUS_"+seed+"((root_M_"+current_seed+"_MINUS_"+seed+"[i-1]+root_M_"+current_seed+"_MINUS_"+seed+"[i])/2)))\n" \
                            "  end if;\n"  \
                            "end do;\n"
                        else:
                            os.error("maple or maxima undefined for \"cas\" parameter")

                        fo_ml.write(dominance_extraction)
                # >>
                print "pareto : ["+str(len(vals))+"] -> ("+str(len(already_seed))+"); trying seed : "+seed
                # print result
                # <<
                if cas == "maxima":
                    result = subprocess.check_output("maxima < "+filename_ml+" | grep -E \"^M[_01]*MINUS[_01]* ## \"", shell=True)
                elif cas == "maple":
                    result = subprocess.check_output("/Applications/Maple\ 2015/maple < "+filename_ml+" | grep -E \"^M[_01]*MINUS[_01]* ## \"", shell=True)
                else:
                    os.error("maple or maxima undefined for \"cas\" parameter")
                    
                if os.path.exists(filename_ml):
                    os.remove(filename_ml)

                print "<"

                # (b.1) Parse the result in one set
                rex_dom  = re.compile('M_([_01]*)_MINUS_([_01]*) ## \[([0-9\.]*)\,([0-9\.]*)\] : ([ 01\-]*)')
                px = rex_dom.findall(result)


                # (b.2.1) first cross to get all the v values
                v = set();
                for p in px:
                    (s1,s2,v1,v2,sg) = p
                    v.add(v1)
                    v.add(v2)
                vl = list(v)
                vl.sort()

                # (b.2.2) second comparison to check if seed is better than any of the best seeds on any given interval
                v_prev = ''
                vl_seed = [None for i in range(len(vl))]
                for i,v_next in enumerate(vl):
                    if v_prev != '':
                        for p in px:
                            (s1,s2,v1,v2,sg) = p
                            # overlap with dominance
                            if (v_prev >= v1 and v_next <= v2):
                                # only set if a current_seed is better than our seed
                                if int(sg) > 0:
                                    vl_seed[i] = s1
                    v_prev = v_next

                # remove first element (by default always None)
                vl_seed = vl_seed[1:]

                # (b.2.3) cross and update the actual pareto if needed
                if None in vl_seed:

                    # (b.2.3.1) merge NONES /or/ similar values into a compact way
                    if len(vl_seed) > 1:
                        if cas == "maxima":
                            vl_new = ['0.0']
                        elif cas == "maple":
                            vl_new = ['0']
                        else:
                            os.error("maple or maxima undefined for \"cas\" parameter")                        
                        vl_seed_new = []
                        for i in range(len(vl_seed)-1):
                            if vl_seed[i] != vl_seed[i+1]:
                                vl_new.append(vl[i+1])
                                vl_seed_new.append(vl_seed[i])
                        if cas == "maxima":
                            vl_new.append('1.0')
                        elif cas == "maple":
                            vl_new.append('1')
                        else:
                            os.error("maple or maxima undefined for \"cas\" parameter")
                        vl_seed_new.append(vl_seed[-1])
                        vl      = vl_new
                        vl_seed = vl_seed_new

                    # (b.2.3..2) pareto update
                    polys[seed] = poly
                    for i,s in enumerate(vl_seed):
                        if s == None:
                            # this means that seed dominate on the interval (vl[i],vl[i+1]) : this must be updated on the pareto
                            vals_new  = []
                            seeds_new = []
                            j = 0
                            # before without overlap
                            while j < len(vals) and vals[j][1] < vl[i]:
                                vals_new.append(vals[j])
                                seeds_new.append(seeds[j])
                                j += 1
                            # partial or total overlap ?
                            if j < len(vals) and vals[j][0] < vl[i]: # partial on left size
                                vals_new.append((vals[j][0],vl[i]))
                                seeds_new.append(seeds[j])
                            vals_new.append((vl[i],vl[i+1]))
                            seeds_new.append(seed)
                            # got to the other size of the None region
                            while j < len(vals) and vals[j][1] <= vl[i+1]:
                                j += 1
                            # partial or total overal ?
                            if j < len(vals) and vals[j][1] > vl[i+1]: # partial on left size
                                vals_new.append((vl[i+1],vals[j][1]))
                                seeds_new.append(seeds[j])
                                j += 1
                            # after without overlap
                            while j < len(vals):
                                vals_new.append(vals[j])
                                seeds_new.append(seeds[j])
                                j += 1
                            # set the new pareto
                            vals  = vals_new
                            seeds = seeds_new

    # Pareto Output
    with open(filename_out, 'w') as fo:
        for i in range(len(vals)):
            print_seedlist([seeds[i]],fo,vals[i][1])


# Main
conds = ["bernoulli","integral","reverse_integral"]
for n in xrange(1,5):
    for w in xrange(3,17):
        for l in xrange(2*w,65):
            for c,cn in enumerate(conds):
                print "_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+".ied",  "_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+"_"+cn+"_python_"+cas+".txt"
                parse_ied_file_condition("experiments/_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+".ied", "experiments/_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+"_"+cn+"_python_"+cas+".txt", l, c)

